/********************************************************************
 *   (C) Copyright OMRON Corporation 1995,1996, all rights reserved *
 ********************************************************************/

#include <stdio.h>
#include "www_wchar.h"

int conv_init = 0;	/* status of initialization */
int conv_d;		/* current designateion */
int conv_mode;		/* conversion mode, ISO-2022 or UNICODE */
int conv_i[4];		/* Invocation of G0,,G3 */
uint32 conv_g[D_LAST];	/* G0,,G3 code sets */
char *conv_bufptr;

wchar_t conv_translate(wchar_t head, unsigned char);
void skip_white_comment();

struct conv_language
    {
    wchar_t head;
    unsigned char *table;
    };
struct conv_language conv_lang[32];

#define IsWhite(c)  (c == ' ' || c == '\n' || c == '\t' || c == '\f' || c == '\r')

/*
 * HTML text to Arena wchar.
 *
 * in: input string.
 * out: output strings by internal wchar.
 * init: 0 = initialize state, 1 = continue.
 * charset: UNICODE or ISO-2022.
 */
int mb2wchar (unsigned char *in, wchar_t *out, int init, char *charset)
{
    static int state;	
    static int translate;
    unsigned char hi, low;

    /* 
     * Hack for Unicode (raw 16bit)
     */
    if (*in == 0)
	{
	init = 1;
	conv_mode = MODE_UNICODE;
	conv_d = D_G0_MB;
	conv_i[I_G0] = D_G0_MB;
	conv_g[D_G0_MB] = HEAD_UNICODE;
	state = S_multi;
	goto init;
	}

    /*
     * decode META tag XXX
     */
    if (init == 0)
	{
	char *start;
	int len = 0;
	conv_bufptr = in;
	/* skip html tag */
	skip_white_comment();
	if (strncasecmp(conv_bufptr, "<html>", 6) == 0)
	    {
	    conv_bufptr += 6;
	    skip_white_comment();
	    }
	/* skip head tag */
	if (strncasecmp(conv_bufptr, "<head>", 6) == 0)
	    {
	    conv_bufptr += 6;
	    skip_white_comment();
	    if (strncasecmp(conv_bufptr, "<meta", 5) == 0)
		{
		conv_bufptr += 5;
		while(IsWhite(*conv_bufptr)) conv_bufptr ++;
		if (strncasecmp(conv_bufptr, "http-equiv", 10) == 0)
		    {
		    conv_bufptr += 10;
		    while(*conv_bufptr == '=' || IsWhite(*conv_bufptr)) conv_bufptr ++;
		    if (strncasecmp(conv_bufptr, "\"Content-Type\"", 14) == 0 ||
			strncasecmp(conv_bufptr, "Content-Type", 12) == 0)
			{
			while(!IsWhite(*conv_bufptr)) conv_bufptr ++;
			while(IsWhite(*conv_bufptr)) conv_bufptr ++;
			if (strncasecmp(conv_bufptr, "content", 7) == 0)
			    {
			    conv_bufptr += 7;
			    while(IsWhite(*conv_bufptr)) conv_bufptr ++;
			    while (*conv_bufptr != '>')
				{
				if (strncasecmp(conv_bufptr, "charset", 7) == 0)
				    {
				    conv_bufptr += 7;
				    while(*conv_bufptr == '=' || IsWhite(*conv_bufptr)) conv_bufptr ++;
				    start = conv_bufptr;
				    while(!IsWhite(*conv_bufptr) && *conv_bufptr != '"') 
					{
					conv_bufptr ++;
					len ++;
					}
				    charset = (char *)malloc(len);
				    strncpy(charset, start, len);
				    }
				conv_bufptr ++;
				}
			    }
			}
		    }
		}
	    }
	else
	    {
	    goto init;
	    }
	}

  init:
    /*
     * Set Unicode or ISO-2022 mode.
     */
    if (init == 0)
	{
	if (strncasecmp(charset, "unicode-1-1-utf-8", 17) == 0)
	    {
	    conv_mode = MODE_UTF8;
	    conv_d = D_G0_MB;
	    conv_i[I_G0] = D_G0_MB;
	    conv_g[D_G0_MB] = HEAD_UNICODE;
	    state = S_single;
	    }
	else if (strncasecmp(charset, "unicode", 7) == 0)
	    {
	    conv_mode = MODE_UNICODE;
	    conv_d = D_G0_MB;
	    conv_i[I_G0] = D_G0_MB;
	    conv_g[D_G0_MB] = HEAD_UNICODE;
	    state = S_multi;
	    }
	else if ((strncasecmp(charset, "x-shift-jis", 11) == 0) ||
		 (strncasecmp(charset, "shift_jis", 9) == 0))
	    {
	    conv_mode = MODE_SHIFT_JIS;
	    conv_d = D_G0_MB;
	    conv_i[I_G0] = D_G0;
	    conv_g[D_G0] = HEAD_ASCII;
	    state = S_single;
	    }
	else
	    {
	    /* 
	     * Initial, designate to G0.
	     * G0: Latin-1
	     * G1: iso-2022-kr
	     */
	    conv_mode = MODE_ISO2022;
	    conv_d = D_G0;
	    conv_i[I_G0] = D_G0;
	    conv_g[D_G0] = HEAD_ASCII;
	    conv_i[I_G1] = D_G1_MB;
	    conv_g[D_G1_MB] = HEAD_KSC5601;
	    translate = 0;
	    state = S_single;
	    }
	}

    /*
     * Main loop
     */
    while ((((conv_mode == MODE_ISO2022) || 
	     (conv_mode == MODE_UTF8) || 
	     (conv_mode == MODE_SHIFT_JIS)) && (*in != NULL)) 
	   || ((conv_mode == MODE_UNICODE) && ((*(in + 1) != NULL) || (*in != NULL))))
	{
	switch (state)
	    {

	    /* single byte code set */
	    case S_single:
	        if (conv_mode == MODE_UTF8)
		    {
		    uint16 ucs2;
		    /* 
		     * Decode UTF8 only in UCS-2
		     */
		    if (*in < 0xC0)
			{
			ucs2 = *in;
			in += 1;
			}
		    else if (*in < 0xE0)
			{
			ucs2 = (*in & 0x3F) << 6 
			    | (*(in + 1) & 0x7F);
			in += 2;
			}
		    else 
			{
			ucs2 = (*in & 0x1F) << 12 
			    | (*(in + 1) & 0x7F) << 6
			    | (*(in + 2) & 0x7F);
			in += 3;
			}
		    *out ++ = ucs2;
		    break;
		    } /* end of MODE_UTF8 */
		if (conv_mode == MODE_SHIFT_JIS)
		    {
		    unsigned int  c1, c2;
		    /*
		     * Decode Shift-JIS
		     */
		    if (*in < 0x80)
			{
			*out ++ = HEAD_ASCII | *in ++;
			}
		    else
			{
			c2 = *in;
			c1 = *(in + 1);
			c2 = c2 + c2 - ((c2 <= 0x9f) ? 0x00e1 : 0x0161);
			if (c1 < 0x9f)
			    {
			    c1 = c1 - ((c1 > 0x7f) ? 0x20 : 0x1f);
			    }
			else {
			    c1 = c1 - 0x7e;
			    c2 ++;
			    }
			hi = c2 & 0xff;
			low = c1 & 0xff;
			*out ++ = HEAD_JISX0208 | (hi << 8) | low;
			in += 2;
			}
		    break;
		    } /* end of MODE_SHIFT_JIS */
	        if (*in == '\033')
		    {
		    state = S_esc;
		    in ++;
		    break;
		    }
		if (*in == SI)
		    {
		    conv_d = conv_i[I_G0];
		    if (conv_d == D_G0_MB)
			{
			state = S_multi;
			}
		    in ++;
		    break;
		    }
		if (*in == SO)
		    {
		    conv_d = conv_i[I_G1];
		    if (conv_d == D_G1_MB)
			{
			state = S_multi;
			}
		    in ++;
		    break;
		    }
		/* 
		 * Translate language specific character
		 * Such as Yen mark, Tilde
		 */
		if ((conv_g[conv_d] == HEAD_ASCII) && translate)
		    {
		    *out++ = conv_translate(translate, *in++);
		    }
		else
		    {
		    *out++ = conv_g[conv_d] | *in ++;
		    }
		break;

	    /* 2bytes code set */
	    case S_multi:
		hi = *in;
		if ((*in == '\033') && (conv_mode != MODE_UNICODE))
		    {
		    state = S_esc;
		    in ++;
		    break;
		    }
		if (*in == SI)
		    {
		    conv_d = conv_i[I_G0];
		    if (conv_d == D_G0)
			{
			state = S_single;
			}
		    in ++;
		    break;
		    }
		if (*in == SO)
		    {
		    conv_d = conv_i[I_G1];
		    if (conv_d == D_G1)
			{
			state = S_single;
			}
		    in ++;
		    break;
		    }
		low = *(in + 1);
		*out ++ = conv_g[conv_d] | (hi << 8) | low;
		in += 2;
		break;

	    case S_esc:
		if (*in == ',')
		    {
		    /* 96, G0 */
		    state = S_comma;
		    in ++;
		    break;
		    }
		if (*in == '-')
		    {
		    /* 96, G1 */
		    state = S_hyphen;
		    in ++;
		    break;
		    }
		if (*in == '.')
		    {
		    /* 96, G2 */
		    state = S_period;
		    in ++;
		    break;
		    }
		if (*in == '(')
		    {
		    /* 94, G0 */
		    state = S_paren_1_l;
		    in ++;
		    break;
		    }
		if (*in == ')')
		    {
		    /* 94, G1 */
		    state = S_paren_1_r;
		    in ++;
		    break;
		    }
		if (*in == '$')
		    {
		    /* 94x94 */
		    state = S_dollar;
		    in ++;
		    break;
		    }
		if (*in == LS2)
		    {
		    /* invoke to G2 */
		    conv_d = conv_i[I_G2];
		    if (conv_d == D_G2_MB)
			{
			state = S_multi;
			}
		    else
			{
			state = S_single;
			}
		    in ++;
		    break;
		    }
		if (*in == LS3)
		    {
		    /* invoke to G3 */
		    conv_d = conv_i[I_G3];
		    if (conv_d == D_G3_MB)
			{
			state = S_multi;
			}
		    else
			{
			state = S_single;
			}
		    in ++;
		    break;
		    }
		/* Single Shift 2 */
		if ((*in & 0x7f) == SS2)
		    {
		    in ++;
		    /* Invoke G2, only single character */
		    if (conv_i[I_G2] == D_G2_MB)
			{
			/* 2byte from G2 */
			hi = *in ++;
			low = *in ++;
			*out ++  = conv_g[conv_i[I_G2]] | (hi << 8) | low;
			}
		    else
			{
			/* 1byte from G2 */
			*out ++ = conv_g[conv_i[I_G2]] | *in ++;
			}
		    /* restore previous state */
		    if (conv_d >= D_MULTI)
			{
			state = S_multi;
			}
		    else
			{
			state = S_single;
			}
		    break;
		    }
		/* Single Shift 3 */
		if ((*in & 0x7f) == SS3)
		    {
		    in ++;
		    /* Invoke G3, only single character */
		    if (conv_i[I_G3] == D_G3_MB)
			{
			/* 2byte from G3 */
			hi = *in ++;
			low = *in ++;
			*out ++  = conv_g[conv_i[I_G3]] | (hi << 8) | low;
			}
		    else
			{
			/* 1byte from G3 */
			*out ++ = conv_g[conv_i[I_G3]] | *in ++;
			}
		    /* restore previous state */
		    if (conv_d >= D_MULTI)
			{
			state = S_multi;
			}
		    else
			{
			state = S_single;
			}
		    break;
		    }
		/* default */
		in ++;
		state = S_single;
		break;

	    /* 96, G1 */
	    case S_hyphen:
		if (conv_d == D_G1_MB)
		    {
		    conv_d = D_G1;
		    }
		if (conv_d >= D_MULTI)
		    {
		    state = S_multi;
		    }
		else
		    {
		    state = S_single;
		    }
		conv_i[I_G1] = D_G1;
		if (*in == 'A')
		    {
		    /* Latin-1 */
		    conv_g[D_G1] = HEAD_ISO8859_1;
		    }
		else
		    {
		    conv_g[D_G1] = HEAD_96 | (*in << 16);
		    }
		in ++;
		break;

	    /* 96, G0. Not specified by ISO 2022, but used by mule */
	    case S_comma:
		if (conv_d == D_G0_MB)
		    {
		    conv_d = D_G0;
		    }
		if (conv_d >= D_MULTI)
		    {
		    state = S_multi;
		    }
		else
		    {
		    state = S_single;
		    }
		conv_i[I_G0] = D_G0;
		if (*in == 'A')
		    {
		    /* Latin-1 */
		    conv_g[D_G0] = HEAD_ISO8859_1;
		    }
		else
		    {
		    conv_g[D_G0] = HEAD_96 | (*in << 16);
		    }
		in ++;
		break;

	    /* 96, G2 */
	    case S_period:
		if (conv_d == D_G2_MB)
		    {
		    conv_d = D_G2;
		    }
		if (conv_d >= D_MULTI)
		    {
		    state = S_multi;
		    }
		else
		    {
		    state = S_single;
		    }
		conv_i[I_G2] = D_G2;
		if (*in == 'A')
		    {
		    /* Latin-1 */
		    conv_g[D_G2] = HEAD_ISO8859_1;
		    }
		else
		    {
		    conv_g[D_G2] = HEAD_96 | (*in << 16);
		    }
		in ++;
		break;

	    case S_dollar:
	    case S_paren_2_l:
		if (*in == '(')
		    {
		    /* 94x94, G0 */
		    state = S_paren_2_l;
		    in ++;
		    break;
		    }
		if (*in == ')')
		    {
		    /* 94x94, G1 */
		    state = S_paren_2_r;
		    in ++;
		    break;
		    }
		if (*in == '*')
		    {
		    /* 94x94, G2 */
		    state = S_star_2;
		    in ++;
		    break;
		    }
		if (*in == '+')
		    {
		    /* 94x94, G3 */
		    state = S_plus_2;
		    in ++;
		    break;
		    }
		if (conv_d == D_G0)
		    {
		    conv_d = D_G0_MB;
		    }
		if (conv_d >= D_MULTI)
		    {
		    state = S_multi;
		    }
		else
		    {
		    state = S_single;
		    }
		conv_i[I_G0] = D_G0_MB;
		conv_g[D_G0_MB] = HEAD_94x94 | (*in << 16);
		in ++;
		break;

	    /* 94x94, G1 */
	    case S_paren_2_r:
		if (conv_d == D_G1)
		    {
		    conv_d = D_G1_MB;
		    }
		if (conv_d >= D_MULTI)
		    {
		    state = S_multi;
		    }
		else
		    {
		    state = S_single;
		    }
		conv_i[I_G1] = D_G1_MB;
		conv_g[D_G1_MB] = HEAD_94x94 | (*in << 16);
		in ++;
		break;

	    /* 94x94, G2 */
	    case S_star_2:
		if (conv_d == D_G2)
		    {
		    conv_d = D_G2_MB;
		    }
		if (conv_d >= D_MULTI)
		    {
		    state = S_multi;
		    }
		else
		    {
		    state = S_single;
		    }
		conv_i[I_G2] = D_G2_MB;
		conv_g[D_G2_MB] = HEAD_94x94 | (*in << 16);
		in ++;
		break;

	    /* 94x94, G3 */
	    case S_plus_2:
		if (conv_d == D_G3)
		    {
		    conv_d = D_G3_MB;
		    }
		if (conv_d >= D_MULTI)
		    {
		    state = S_multi;
		    }
		else
		    {
		    state = S_single;
		    }
		conv_i[I_G3] = D_G3_MB;
		conv_g[D_G3_MB] = HEAD_94x94 | (*in << 16);
		in ++;
		break;

	    /* 94, G0 */
	    case S_paren_1_l:
		if (conv_d == D_G0_MB)
		    {
		    conv_d = D_G0;
		    }
		if (conv_d >= D_MULTI)
		    {
		    state = S_multi;
		    }
		else
		    {
		    state = S_single;
		    }
		conv_i[I_G0] = D_G0;
		if (*in == 'B')
		    {
		    conv_g[D_G0] = HEAD_ASCII;
		    translate = 0;
		    in ++;
		    break;
		    }
		if (*in == 'J')
		    {
		    conv_g[D_G0] = HEAD_ASCII;
		    translate = HEAD_JIS_ROMAN;
		    in ++;
		    break;
		    }
		if (*in == 'T')
		    {
		    conv_g[D_G0] = HEAD_ASCII;
		    translate = HEAD_GB_ROMAN;
		    in ++;
		    break;
		    }
		conv_g[D_G0] = HEAD_94 | (*in << 16);
		in ++;
		break;

	    /* 94, G1 */
	    case S_paren_1_r:
		if (conv_d == D_G1_MB)
		    {
		    conv_d = D_G1;
		    }
		if (conv_d >= D_MULTI)
		    {
		    state = S_multi;
		    }
		else
		    {
		    state = S_single;
		    }
		conv_i[I_G1] = D_G1;
		conv_g[D_G1] = HEAD_94 | (*in << 16);
		in ++;
		break;

	    }
	}
    *out = 0;	/* output terminate by 0 */
}

/*
 * ASCII variant (JIS-Roman, GB-Roman) convert to ASCII and
 * uniq character.
 *
 * head:	Language header. (internal wchar header)
 * ch:   	Translate character. (single byte)
 *
 *       conv_lang[]
 *  +----------+-------+
 *  | head     | table |         ch
 *  +----------+-------+      This code must be original language. 
 *  | 614a0000 |   *---|----->+------+
 *  | 61540000 |       |      | 0x5c |   
 *  |        0 |       |      | 0x7c |
 * 			      | 0x7e |
 *                            |    0 |
 *			      +------+
 *
 */ 

wchar_t conv_translate(wchar_t head, unsigned char ch)
{
int t,c;
unsigned char *table = NULL;

/* initialize conversion table */
if (conv_init == 0)
    {
    conv_init_translate();
    conv_init = 1;
    }

/* look for translate table */
for (t = 0; t < 32; t ++)
    {
    if (conv_lang[t].head == head)
	{
	table = conv_lang[t].table;
	break;
	}
    if (conv_lang[t].head == 0)
	{
	table = NULL;
	break;
	}
    }

/* use default table */
if (table == NULL)
    {
    return((wchar_t)ch);
    }

/* translate by specific table */
for(c = 0; c < 64; c ++)
    {
    if (table[c] == ch)
	{
	return (head | ch);
	}
    if (table[c] == 0)
	{
	return ((wchar_t)ch);
	}
    }
}  

conv_init_translate()
{
unsigned char *table;

/* JIS Roman */
conv_lang[0].head = HEAD_JIS_ROMAN;
conv_lang[0].table = (unsigned char *)malloc(sizeof(unsigned char) * 4);
table = conv_lang[0].table;

table[0] = 0x5c;	/* backslash	--> yen symbol */
table[1] = 0x7c;	/* broken bar	--> bar */
table[2] = 0x7e;	/* tilde 	--> overbar */
table[3] = 0;		/* terminate */

/* GB Roman */
conv_lang[1].head = HEAD_GB_ROMAN;
conv_lang[1].table = (unsigned char *)malloc(sizeof(unsigned char) * 2);
table = conv_lang[1].table;

table[0] = 0x24;	/* dollar sign 	--> Chinese Yen */
table[1] = 0;		/* terminate */


/* Termination */
conv_lang[2].head = 0;
conv_lang[2].table = NULL;

}


/*
 * Skip HTML comment and WHITESPACES
 */
void skip_white_comment()
{
    /* skip WHITESPACE */
    while(IsWhite(*conv_bufptr)) *conv_bufptr ++;
    /* find HTML comment tag */
    if (strncasecmp(conv_bufptr, "<!--", 4) == 0)
	{
	conv_bufptr += 4;
	while (*conv_bufptr != NULL)
	    {
	    /* find end of HTML comment tag */
	    if (strncasecmp(conv_bufptr, "-->", 3) == 0)
		{
		conv_bufptr += 3;
		/* skip WHITESPACE */
		while(IsWhite(*conv_bufptr)) conv_bufptr ++;
		return;
		}
	    conv_bufptr ++;
	    }
	}
}


/* 
 *  Check Kinsoku character for Japanese
 */
int kinsoku(c)
int c;
{
    int rv;
    rv = kinsoku_ja(c) || kinsoku_cn(c) || kinsoku_kr(c);
    return rv;
}

int kinsoku_ja(c)
int c;
{
    int head, body;
    head = c & HEAD_MASK;
    body = c & 0x0000ffff;

    if (head == HEAD_JISX0208 || 
	head == HEAD_JISX0208_O)
	{
	if (((body >= 0x2122) && (body <= 0x2145)) ||
	    body == 0x214b || /* maru kokka */
	    body == 0x214d || /* kaku kokka */
	    body == 0x214f || /* Kaku kokka */
	    body == 0x2151 || /* nyoro kokka */
	    body == 0x2153 || /* lt kokka */
	    body == 0x2155 || /* LT kokka */
	    body == 0x2157 || /* kagi kokka */
	    body == 0x2159 || /* Kagi kokka */
	    body == 0x215b || /* kakumaru kokka */
	    body == 0x216b || /* Degree */
	    body == 0x216c || /* Dash */
	    body == 0x216d || /* Double dash */
	    body == 0x216e || /* Cerusius */
	    body == 0x2421 || /* Small Romaji "a" */
	    body == 0x2423 || /* Small Romaji "i" */
	    body == 0x2425 || /* Small Romaji "u" */
	    body == 0x2427 || /* Small Romaji "e" */
	    body == 0x2429 || /* Small Romaji "o" */
	    body == 0x2443 || /* Small Romaji "tsu" */
	    body == 0x2463 || /* Small Romaji "ya" */
	    body == 0x2465 || /* Small Romaji "yu" */
	    body == 0x2467 || /* Small Romaji "yo" */
	    body == 0x246e || /* Small Romaji "wa" */
	    body == 0x2521 || /* Small Romaji "A" */
	    body == 0x2523 || /* Small Romaji "I" */
	    body == 0x2525 || /* Small Romaji "U" */
	    body == 0x2527 || /* Small Romaji "E" */
	    body == 0x2529 || /* Small Romaji "O" */
	    body == 0x2543 || /* Small Romaji "Tsu" */
	    body == 0x2563 || /* Small Romaji "Ya" */
	    body == 0x2565 || /* Small Romaji "Yu" */
	    body == 0x2567 || /* Small Romaji "Yo" */
	    body == 0x256e || /* Small Romaji "Wa" */
	    body == 0x2575 || /* Small Romaji "Ka" */
	    body == 0x2576)   /* Small Romaji "Ke" */
	    {
	    return (1);
	    }
	else
	    {
	    return (0);
	    }
	}
}

int kinsoku_cn(c)
int c;
{
    return(0);
}

int kinsoku_kr(c)
int c;
{
    return(0);
}
