/********************************************************************
 *   (C) Copyright OMRON Corporation 1995,1996, all rights reserved *
 ********************************************************************/

/* ad hoc text drawing functions for X11 */

#include <X11/Xlib.h>

/* XXX: for debugging */
extern int debug;
#define	FONT_TRACE	(debug & 0x8000)
extern int NoStyle;

#include "www_wchar.h"
#include "text_x11.h"

#ifndef   NULL
#define NULL 0
#endif /* NULL */

extern void* malloc(int);		/* <stdlib.h> */
extern void* strdup(void*);		/* <string.h> */

#define	BUF_SIZE	512

/* Drawing routines
 *	GxDrawString()
 *	GxDrawImageString()
 */

int GxDrawString(Display *display, Drawable drawable, OutC outc, GC gc,
		 int x, int y, char *string, int length)
{
    XGCValues gcval;
    OutCStat stat;
    char buf[BUF_SIZE];
    int  len;
    XFontStruct *fs;
    int escapement;

    XGetGCValues(display, gc, GCFont, &gcval);

    escapement = 0;
    stat = (*outc->init_conv)(string, length);
    while ((len = (*outc->do_conv)(outc, stat, buf, BUF_SIZE, &fs)) > 0) {
	XSetFont(display, gc, fs->fid);
	if (fs->min_byte1 == 0 && fs->max_byte1 == 0) {
	    /* one-byte font */
	    XDrawString(display, drawable, gc, x + escapement, y, buf, len);
	    escapement += XTextWidth(fs, buf, len);
	} else {
	    /* two-byte font */
	    XDrawString16(display, drawable, gc, x + escapement, y,
			  (XChar2b*)buf, len/2);
	    escapement += XTextWidth16(fs, (XChar2b*)buf, len/2);
	}
    }
    (*outc->term_conv)(stat);

    XChangeGC(display, gc, GCFont, &gcval);
    return escapement;
}

int GxDrawImageString(Display *display, Drawable drawable, OutC outc, GC gc,
		      int x, int y, char *string, int length)
{
    XGCValues gcval;
    OutCStat stat;
    char buf[BUF_SIZE];
    int  len;
    XFontStruct *fs;
    int escapement;

    XGetGCValues(display, gc, GCFont, &gcval);

    escapement = 0;
    stat = (*outc->init_conv)(string, length);
    while ((len = (*outc->do_conv)(outc, stat, buf, BUF_SIZE, &fs)) > 0) {
	XSetFont(display, gc, fs->fid);
	if (fs->min_byte1 == 0 && fs->max_byte1 == 0) {
	    /* one-byte font */
	    XDrawImageString(display, drawable, gc, x + escapement, y,
			     buf, len);
	    escapement += XTextWidth(fs, buf, len);
	} else {
	    /* two-byte font */
	    XDrawImageString16(display, drawable, gc, x + escapement, y,
			       (XChar2b*)buf, len/2);
	    escapement += XTextWidth16(fs, (XChar2b*)buf, len/2);
	}
    }
    (*outc->term_conv)(stat);

    XChangeGC(display, gc, GCFont, &gcval);
    return escapement;
}


/* Text Escapement and Extents
 *	GxTextEscapement()
 */

int GxTextEscapement(OutC outc, char *string, int count)
{
    XGCValues gcval;
    OutCStat stat;
    char buf[BUF_SIZE];
    int  len;
    XFontStruct *fs;
    int escapement;
    static XFontStruct *old_fs;
    static int old_escapement;

    escapement = 0;
    stat = (*outc->init_conv)(string, count);
    while ((len = (*outc->do_conv)(outc, stat, buf, BUF_SIZE, &fs)) > 0) {
	if (fs->min_byte1 == 0 && fs->max_byte1 == 0) {
	    /* one-byte font */
	    escapement += XTextWidth(fs, buf, len);
	    old_fs = fs;
	} else {
	    /* two-byte font */
	    if ((fs == old_fs) && (len == 2))
		{
		/* special case for word wraping */
		escapement += old_escapement;
		}
	    else
		{
		old_fs = fs;
		old_escapement = XTextWidth16(fs, (XChar2b*)buf, len/2);
		escapement += old_escapement;
		if (len != 2)
		    {
		    old_fs = NULL;
		    }
		}
	}
    }
    (*outc->term_conv)(stat);
    return escapement;
}


/* generic converter functions for wchar_t
 *	wc_init_conv()	initializer for generic wchar_t
 *	wc_do_conv()	cutdown for generic wchar_t
 *	wc_term_conv()
 */
static int dbg = 0;
debug_conv(int level) { dbg = level; }

struct _OutCStatRec {
    wchar_t	*wstring;
    int		length;		/* size in wchar_t */
    int		cur;		/* index in wchar_t */
};

static OutCStat wc_init_conv(char *string, int length)
{
    OutCStat stat;

    stat = (OutCStat)malloc(sizeof(OutCStatRec));
    if (stat) {
#ifndef   SOLARIS
	bzero(stat, sizeof(OutCStatRec));
#else  /* SOLARIS */
	memset(stat, 0, sizeof(OutCStatRec));
#endif /* SOLARIS */
	stat->wstring = (wchar_t*)string;
	stat->length = length / sizeof(wchar_t);
	stat->cur    = 0;
    }
    return stat;
}

static void wc_term_conv(OutCStat stat)
{
    free(stat);
}

static int wc_do_conv(OutC outc, OutCStat stat,
		      char *buf, int buflen, XFontStruct **fs)
{
    wchar_t *wstr;
    int iw, is;
    int codeset, code;
    int graph, nbytes;
    int mask, msb;
    int (*post_proc)();

    if (dbg)
	printf("wc_do_conv(outc=%08x,stat=%08x,buf=%08x,buflen=%d,fs=%08x)\n",
	       outc, stat, buf, buflen, fs);
    wstr = stat->wstring;
    is = 0;
    codeset = -1;
    post_proc = NULL;
    for (iw = stat->cur; iw < stat->length; iw++) {
	if (dbg)
	    printf("wc_do_conv() wstr[%d]=%08x\n", iw, wstr[iw]);
	if (codeset == -1) {  /* first time */
	    codeset = (*outc->getcode)(outc, wstr[iw], &code, &nbytes);
	    if ((*outc->query_font)(outc, codeset, fs, &graph, &post_proc)) {
		switch (graph) {
		 case 2:  mask = 0xff; msb = 0x00; break;
		 case 1:  mask = 0xff; msb = 0x80; break;
		 case 0:  mask = 0x7f; msb = 0x00; break;
		}
	    } else {
		nbytes = -1;	/* no font is available */
	    }
	} else if ((*outc->getcode)(outc, wstr[iw], &code, NULL) != codeset) {
	    /* different codeset starts */
	    break;
	}
	switch (nbytes) {
	 case 1:
	    buflen -= 1;
	    if (buflen <= 0)
		break;
	    buf[is++] = ( code       & mask) | msb;
	    break;
	 case 2:
	    buflen -= 2;
	    if (buflen <= 0)
		break;
	    buf[is++] = ((code >> 8) & mask) | msb;
	    buf[is++] = ( code       & mask) | msb;
	    break;
	 default: /* codeset is invalid or no font is availble */
	    buflen -= 10; /*XXX*/
	    if (buflen <= 0)
		break;
	    sprintf(&buf[is], "[%08x]", wstr[iw]);
	    is += 10; /*XXX*/
	    break;
	}
    }
    buf[is] = '\0';
    stat->cur = iw;
    if (post_proc)
	(*post_proc)(buf, is);

    if (dbg)
	printf("wc_do_conv() buf=\"%s\"(len:%d,%d)\n", buf, buflen, is);

    return is;
}


/* specific functions for wchar_t designed by shin@nff.ncl.omron.co.jp */
static int awc_getcode(OutC outc, wchar_t wc, int *code, int *nbytes)
{
    int codeset;
    char byte1, byte2, byte3, byte4;

    /* At first, process special cases */
    if (IS_LATIN_1(wc)) {
	if (code)
	    *code = wc & 0xff;
	if (nbytes)
	    *nbytes = 1;
	return HEAD_ISO8859_1;
    }
    if (IS_UNICODE(wc)) {
	/* XXX: convert to ordinally codes */
	if (code)
	    *code = wc & 0xffff;
	if (nbytes)
	    *nbytes = 2;
	return HEAD_UNICODE;
    }

    byte1 = (wc >> 24) & 0xff;
    byte2 = (wc >> 16) & 0xff;
    byte3 = (wc >>  8) & 0xff;
    byte4 =  wc        & 0xff;
    switch (byte1) {
     case 0x61: /* 94 */
     case 0x62: /* 96 */
	if (!isprint(byte2) || byte3)
	    break;
	if (code)
	    *code = wc & 0xff;
	if (nbytes)
	    *nbytes = 1;
	return wc & HEAD_MASK;
     case 0x63: /* 94x94 */
     case 0x64: /* 96x96 */
	if (!isprint(byte2))
	    break;
	if (code)
	    *code = wc & 0xffff;
	if (nbytes)
	    *nbytes = 2;
	return wc & HEAD_MASK;
    }

    if (code)
        *code = -1;
    if (nbytes)
        *nbytes = -1;
    return -1;
}

/***********************************************************************
 *
 *	Font handling
 *
 */

static XFontStruct* load_font(OutCList, char*);
static XFontStruct* try_font(char*, char*);
static XFontStruct* try_font2(char*);

static int reverse(char* buf, int buflen)	/* XXX: for RtoL,1byte */
{
    int s, t;
    char tmp;

    for (s = 0, t = buflen-1; s < t; s++, t--) {
	tmp    = buf[s];
	buf[s] = buf[t];
	buf[t] = tmp;
    }
    return 1;
}

static struct {
    int		codeset;
    int		graphic;
    int		width;			/* width in preformated mode */
    int		(*post_proc)();
    char	*charset;
    char	*fontname;
    XFontStruct	*fs;
} fontinfo[] = {
    /* See http://133.210.2.244/demo/multi-sample.html */
    { HEAD_JISX0201,	/* 61490000: Japanese Katakana */
	  1, 1,	NULL,		"jisx0201.1976-0",
	  "-*-fixed-medium-r-normal--14-130-75-75-c-70-jisx0201.1976-0",},
    { 0x614a0000,	/* 614a0000: Japanese Roman */
	  0, 1,	NULL,		"jisx0201.1976-0",
	  "-*-fixed-medium-r-normal--14-130-75-75-c-70-jisx0201.1976-0",},
    { 0x62310000,	/* 62310000: viscii */
	  1, 1,	NULL,		"viscii1.1-1",
	  "-etl-fixed-medium-r-normal--16-160-72-72-c-80-viscii1.1-1", },
    { HEAD_ISO8859_1,	/* 62410000: Latin-1 */
	  2, 1,	NULL,		"iso8859-1",
	  "-etl-fixed-medium-r-normal--14-140-72-72-c-70-iso8859-1", },
    { HEAD_ISO8859_2,	/* 62420000: Latin-2 */
	  1, 1,	NULL,		"iso8859-2",
	  "-etl-fixed-medium-r-normal--14-140-72-72-c-70-iso8859-2", },
    { HEAD_ISO8859_3,	/* 62430000: Latin-3 */
	  1, 1,	NULL,		"iso8859-3",
	  "-etl-fixed-medium-r-normal--14-140-72-72-c-70-iso8859-3", },
    { HEAD_ISO8859_4,	/* 62440000: Latin-4 */
	  1, 1,	NULL,		"iso8859-4",
	  "-etl-fixed-medium-r-normal--14-140-72-72-c-70-iso8859-4", },
    { HEAD_ISO8859_5,	/* 624C0000: Cyrillic */
	  1, 1,	NULL,		"iso8859-5",
	  "-etl-fixed-medium-r-normal--14-140-72-72-c-70-iso8859-5", },
    { HEAD_ISO8859_7,	/* 62460000: Greek */
	  1, 1,	NULL,		"iso8859-7",
	  "-etl-fixed-medium-r-normal--14-140-72-72-c-70-iso8859-7", },
    { HEAD_ISO8859_8,	/* 62480000: Hebrew */
	  1, 1,	reverse,	"iso8859-8",
	  "-etl-fixed-medium-r-normal--14-140-72-72-c-70-iso8859-8", },
    { HEAD_ISO8859_9,	/* 624D0000: Latin-5 */
	  1, 1,	NULL,		"iso8859-9",
	  "-etl-fixed-medium-r-normal--14-140-72-72-c-70-iso8859-9", },
    { 0x62540000,	/* 62540000: Thai */
	  1, 1,	NULL,		"tis620.2529-1",
	  "-etl-fixed-medium-r-normal--14-140-72-72-m-70-tis620.2529-1", },
    { 0x63400000,	/* 63400000: XXX Japanese Kanji */
	  0, 2,	NULL,		"jisx0208.1983-0",
	  "-jis-fixed-medium-r-normal--0-0-75-75-c-0-jisx0208.1983-0", },
    { HEAD_GB2312,	/* 63410000: Chainese Hanji */
	  0, 2,	NULL,		"gb2312.1980-0",
	  "-*-song ti-medium-r-normal--16-160-72-72-c-160-gb2312.1980-0", },
    { HEAD_JISX0208,	/* 63420000: Japanese Kanji */
	  0, 2,	NULL,		"jisx0208.1983-0",
	  "-misc-fixed-medium-r-normal--14-130-75-75-c-140-jisx0208.1983-0", },
    { HEAD_KSC5601,	/* 63430000: Korean Hanja */
	  0, 2,	NULL,		"ksc5601.1987-0",
	  "-*-mincho-medium-r-normal--16-120-100-100-c-160-ksc5601.1987-0", },
    { HEAD_JISX0212,	/* 63440000: New Japanese Kanji */
	  0, 2,	NULL,		"jisx0212.1990-0",
	  "-misc-fixed-medium-r-normal--16-150-75-75-c-160-jisx0212.1990-0", },
    { HEAD_CNS11643_1,	/* 63470000: Complexed chinese 1 */
	  0, 2,	NULL,		"cns11643.1-0",
	  "-cbs-song-medium-r-normal--16-160-72-72-c-160-cns11643.1-0", },
    { HEAD_CNS11643_2,	/* 63480000: Complexed chinese 2 */
	  0, 2,	NULL,		"cns11643.2-0",
	  "-cbs-song-medium-r-normal--16-160-72-72-c-160-cns11643.2-0", },
    { HEAD_UNICODE,	/* 00000000: Unicode-1-1 */
	  2, 2,	NULL,		"unicode1.1-0",
	  "-ccic-soung-midium-r-normal--24-240-75-75-c-240-unicode1.1-0", },
#ifdef BIG5
    { 0x63300000,	/* 63300000: BIG5? */
	  2, 2,	NULL,		"big5-0",
	  "-big5-cclib-medium-r-normal-fs-24-230-75-75-c-240-big5-0", },
#endif
    { 0, 0, 0,	NULL,	NULL,	NULL, }
};

static int query_font(OutC outc, int codeset,
		      XFontStruct **fs, int *graph, int (**proc)())
{
    extern Display *display;	/* XXX */
    XFontStruct *fonts;
    int i;

    /* Look up a font table */
    for (i = 0; fontinfo[i].charset; i++) {
	if (codeset == fontinfo[i].codeset) {
	    if (codeset == HEAD_ISO8859_1) {
		/* XXX: This is a special case now */
		/* XXX: should query to font DB */
		fonts = outc->default_fonts;
	    } else {
		if (NoStyle) {
		    /* XXX: doesn't use style sheet information */
		    if (!fontinfo[i].fs) {
			fontinfo[i].fs = XLoadQueryFont(display,
							fontinfo[i].fontname);
		    }
		    fonts = fontinfo[i].fs;
		} else {
		    /* XXX: should check font cache of each outc */
		    fonts = load_font(outc->xlfd_list, fontinfo[i].charset);
		}
	    }
	    if (fonts) {
		if (fs)
		    *fs = fonts;
		if (graph)
		    *graph = fontinfo[i].graphic;
		if (proc)
		    *proc = fontinfo[i].post_proc;
		return 1;
	    }
	    break;
	}
    }

    /* no font is available for the codeset */
    if (fs)
        *fs = (XFontStruct*) outc->default_fonts;		/* XXX */
    if (graph)
        *graph = -1;
    if (proc)
        *proc = NULL;
    return 0;
}

OutC CreateOutC(OutCList xlfd_list)
{
    OutC outc;
    XFontStruct *fs;

    outc = (OutC)malloc(sizeof(OutCRec));
    if (outc == NULL)
	return (OutC)NULL;

    fs = load_font(xlfd_list, "iso8859-1");	/* XXX */
    if (fs == NULL) {
	free(outc);
	return (OutC)NULL;
    }

    outc->init_conv	= wc_init_conv;		/* XXX */
    outc->do_conv	= wc_do_conv;		/* XXX */
    outc->term_conv	= wc_term_conv;		/* XXX */
    outc->getcode	= awc_getcode;		/* XXX */
    outc->query_font	= query_font;
    outc->default_fonts	= fs;			/* XXX */
    outc->xlfd_list	= xlfd_list;

    return outc;
}

XFontStruct *DefaultFontOfOutC(OutC outc)
{
    return outc->default_fonts;
}

struct _OutCListRec {
    void*		object;
    struct _OutCListRec	*next;
};

OutCList add_list(OutCList list, void* object)
{
    OutCList newl, tmpl;

    if (object == NULL)
	return list;

    newl = (OutCList)malloc(sizeof(OutCListRec));
    newl->object = object;
    newl->next   = NULL;

    if (list == NULL)
	return newl;

    for (tmpl = list; tmpl->next; tmpl = tmpl->next) {
    }
    tmpl->next = newl;
    return list;
}    


static XFontStruct *load_font(OutCList xlfd_list, char* charset)
{
    OutCList xlfd;
    XFontStruct *fs = NULL;

    for (xlfd = xlfd_list; xlfd; xlfd = xlfd->next) {
	fs = try_font(xlfd->object, charset);
	if (fs)
	    break;
    }
    return fs;
}

static XFontStruct *try_font(char* prefix, char* charset)
{
    char xlfd[256];
    int i;
    int len;

    if (FONT_TRACE)
	printf("try_font(prefix=%s, charset=%s)\n", prefix, charset);

    len = strlen(prefix);

    /* XXX: process a special case "symbol" */
    if (strcmp(charset, "iso8859-1") == 0) {
	for (i = 1; i < len; i++) {
	    if (prefix[i] == '-') {
		if ((prefix[i+1] == 's' || prefix[i+1] == 'S')
		 && strncasecmp(&prefix[i+1], "symbol", 6) == 0) {
		    charset = "*";	/* XXX */
		}
		break;
	    }
	}
    }

    /* XXX: ad hoc rewriting */
    strcpy(xlfd, prefix);
    if (strcmp(&xlfd[len-4], "-*-*") == 0) {
	strcpy(&xlfd[len-3], charset);
    }

    /* XXX: you should choose appropriate `pixel' field of XLFD,
     * before calling try_font2() */
    return try_font2(xlfd);
}

struct loaded_font {
    char		*xlfd;		/* XLFD			*/
    XFontStruct		*fs;		/* XFontStruct		*/
    int			refcnt;		/* reference count	*/
    struct loaded_font	*next;		/* next object		*/
} *loaded_font_list = NULL;

static XFontStruct *try_font2(char* xlfd)
{
    extern Display *display;	/* XXX */
    XFontStruct *fs;
    int i;
    struct loaded_font *lf;
    
    /* search from DB */
    for (lf = loaded_font_list; lf; lf = lf->next) {
	if (strcmp(xlfd, lf->xlfd) == 0) {
	    lf->refcnt++;
	    if (FONT_TRACE)
		printf("try_font2(%s) found font in cache (refcnt=%d)\n",
		       xlfd, lf->refcnt);
	    return lf->fs;
	}
    }

    /* load it */
    fs = XLoadQueryFont(display, xlfd);
    if (fs == NULL) {
	if (FONT_TRACE)
	    printf("try_font2(%s) failed!\n", xlfd);
	return NULL;
    }

    lf = (struct loaded_font*)malloc(sizeof(struct loaded_font));
    if (lf == NULL) {
	XFreeFont(display, fs);
	return NULL;
    }
    lf->xlfd   = strdup(xlfd);
    lf->fs     = fs;
    lf->refcnt = 1;
    lf->next   = loaded_font_list;
    loaded_font_list = lf;
    if (FONT_TRACE)
	printf("try_font2(%s) succeeded!\n", xlfd);
    return lf->fs;
}
