/********************************************************************
 *   (C) Copyright OMRON Corporation 1995,1996, all rights reserved *
 ********************************************************************/

/*
 * define basic types
 */

typedef unsigned int	uint32;
typedef unsigned short	uint16;
typedef unsigned char	uint8;
typedef uint32	CCL_wchar;

#define wchar_t	CCL_wchar



#define	MASK_UNICODE	((uint32)0xFFFF0000)
#define MASK_LATIN_1	((uint32)0xFFFFFF00)
#define MASK_ASCII	((uint32)0xFFFFFF80)



#define IS_UNICODE(c)	(((c) & MASK_UNICODE) == 0)
#define IS_LATIN_1(c)	(((c) & MASK_LATIN_1) == 0)
#define IS_ASCII(c)	(((c) & MASK_ASCII) == 0)

#define TO_ASCII(c)	((char)(c))

#define HEAD_MASK	((uint32)0xFFFF0000)

#define HEAD_UNICODE	((uint32)0x00000000)	/* unicode-1-1 */
#define HEAD_ASCII      ((uint32)0x00000000)	/* ascii */
#define HEAD_JIS_ROMAN  ((uint32)0x614a0000)	/* JIS X 0201-1976 Roman */
#define HEAD_GB_ROMAN   ((uint32)0x61540000)	/* GB 1988-89 Roman */
#define HEAD_ISO8859_1	((uint32)0x00000080)	/* Latin-1 */
#define HEAD_ISO8859_2	((uint32)0x62420000)	/* Latin-2 */
#define HEAD_ISO8859_3	((uint32)0x62430000)	/* Latin-3 */
#define HEAD_ISO8859_4	((uint32)0x62440000)	/* Latin-4 */
#define HEAD_ISO8859_5	((uint32)0x624C0000)	/* Cyrillic */
#define HEAD_ISO8859_6	((uint32)0x62470000)	/* Arabic */
#define HEAD_ISO8859_7	((uint32)0x62460000)	/* Greek */
#define HEAD_ISO8859_8	((uint32)0x62480000)	/* Hevrew */
#define HEAD_ISO8859_9	((uint32)0x624D0000)	/* Latin-5 */
#define HEAD_ISO8859_10	((uint32)0x62560000)	/* Latin-6 */

#define HEAD_JISX0208_O	((uint32)0x63400000)	/* Japanese Kanji (1978) */
#define HEAD_GB2312	((uint32)0x63410000)	/* Chinese Kanji */
#define HEAD_JISX0208	((uint32)0x63420000)	/* Japanese Kanji (1983) */
#define HEAD_JISX0201	((uint32)0x61490000)	/* Katakana */
#define HEAD_KSC5601	((uint32)0x63430000)	/* Korean */
#define HEAD_JISX0212	((uint32)0x63440000)	/* New JIS */
#define HEAD_CNS11643_1 ((uint32)0x63470000)
#define HEAD_CNS11643_2 ((uint32)0x63480000)


#define HEAD_94		((uint32)0x61000000)
#define HEAD_96		((uint32)0x62000000)
#define HEAD_94x94	((uint32)0x63000000)
#define HEAD_96x96	((uint32)0x64000000)


/* state of escape sequece */

#define S_ascii		0
#define S_nonascii	1
#define S_esc		2
#define S_hyphen	3
#define S_dollar	4
#define S_paren_1_l	5
#define S_paren_1_r	6
#define S_comma		7
#define S_single	8
#define S_multi		9
#define S_paren_2_l	10
#define S_paren_2_r	11
#define S_period	12
#define S_star_2	13
#define S_plus_2	14

/* current designation */

#define D_G0	 	 0
#define D_G0_MB		10
#define D_G1		 1
#define D_G1_MB		11
#define D_G2		 2
#define D_G2_MB		12
#define D_G3		 3
#define D_G3_MB		13
#define D_LAST		14
#define D_MULTI		10

#define I_G0		 0
#define I_G1		 1
#define I_G2		 2
#define I_G3		 3


/* code converter mode */
#define MODE_UNICODE	0
#define MODE_ISO2022	1
#define MODE_UTF8	2
#define MODE_SHIFT_JIS	3
#define MODE_BIG5	4

#define SI	0x0F	/* 017 */
#define SO	0x0E	/* 016 */
#define LS2	0x6E
#define LS3	0x6F
#define SS2	0x4E
#define SS3	0x4F


/*
 Escape sequece state diagram


SI  =	designate `G0'
SO  =	designate `G1'
LS2 =   designate `G2'
LS3 =   designate `G3'

ESC
    --> ","	--> "A -- V"	= 96(G0)
    --> "-"	--> "A -- V"	= 96(G1)
    --> "."	--> "A -- V"	= 96(G2)

    --> "("	--> "B"		= 94(G0), ASCII
		--> "J"		= 94(G0), JIS Roman
		--> "I"		= 94(G0), JIS Katakana

    --> ")"	--> "A -- V"	= 94(G1), ASCII

    --> "$"	--> "@AB"	= 94x94(G0)

    		--> "("		--> "A -- H"	= 94x94(G0)
		--> ")"		--> "A -- H"	= 94x94(G1)

   --> 0x6E	= LS2(G2)
   --> 0x6F	= LS3(G3)

*/

#ifdef    SUNOS
#define size_t unsigned int
#endif /* SUNOS */

/* proto types for wutil.c */
wchar_t *W_strstr(wchar_t *s, char *find);
wchar_t W_toupper(wchar_t c);
wchar_t W_tolower(wchar_t c);
char *W_strndup(wchar_t *s, int n);
wchar_t *WW_strndup(wchar_t *s, int len);
char *W_strdup(wchar_t *s);
wchar_t *WW_strdup(wchar_t *s);
wchar_t *w_strdup(char *s);
wchar_t *w_strndup(char *s, int len);
int W_strcmp(wchar_t *s1, char *s2);
int WW_strcmp(wchar_t *s1, wchar_t *s2);
int W_strncmp(wchar_t *s1, char *s2, size_t n);
int WW_strncmp(wchar_t *s1, wchar_t *s2, size_t n);
int W_strcasecmp(wchar_t *s1, char *s2);
int W_strncasecmp(wchar_t *s1, unsigned char * s2, int len);
size_t W_strlen(wchar_t *str);
wchar_t *W_strchr(wchar_t *str, char c);
int W_atoi(wchar_t *str);
int STR_NCASECMP(wchar_t *, wchar_t *, int);
wchar_t *WW_strncpy(wchar_t *, wchar_t *, size_t);
char *W_strncpy(char *, wchar_t *, size_t);
