#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>

#include "www.h"

extern Display *display;
extern Colormap colormap;
extern double lens_factor;
extern int screen;

extern XFontStruct *Fonts[FONTS];  /* array of fonts */
extern char *FontsName[FONTS];
extern int LineSpacing[FONTS];
extern int BaseLine[FONTS];
extern int StrikeLine[FONTS];
extern int LineThickness[FONTS];
extern int debug;

long ParseColor(const char *s)
{
    XColor xc;

    if (XParseColor(display, colormap, s, &xc))
	return (((xc.red >> 8) << 16) | ((xc.green >> 8) << 8) | (xc.blue >> 8));
    return 0; /* i.e. black */
}


#ifndef I18N

int myLoadFont(char *name)
{
    int direction_hint, font_ascent, font_descent;
    XCharStruct overall;
    XFontStruct *font = NULL;
    static int fix = 0;
    /* janet 24/07/95: not used:    int use_fix = -1; */
    int i;
    char *test = "Testing";
    
    for(i=0; i < fix; i++) {
	if (strcmp(name, FontsName[i]) == 0) {
	    if (FONT_TRACE)
		fprintf(stderr,"font: %s found in cache, fix = %d\n",name, i);
    	    return i;
	} 
    }

    /* a font with the same name has not been loaded, so try it.. */

    font = XLoadQueryFont(display, name);
    
    if (font) {
	FontsName[fix] = strdup(name);
	Fonts[fix] = font;

	XTextExtents(font, test, strlen(test),
		     &direction_hint,
		     &font_ascent,
		     &font_descent,
		     &overall);

	LineSpacing[fix] = SPACING(font);
	BaseLine[fix] = BASELINE(font);
	StrikeLine[fix] = STRIKELINE(font);
	LineThickness[fix] = LINETHICKNESS(font);
	fix++;
/*	fprintf(stderr,"FIX %d\n",fix);*/
	if (FONT_TRACE)
	    fprintf(stderr,"font %s: xloadqueryfonts successful\n",name);
	return (fix - 1);
    }
    
    if (FONT_TRACE)
	fprintf(stderr,"font: mloadFont failed %s\n",name);
    return (0); /* default font returned on error */
}


int GetFont(char *family_l, long px, long weight, long style, BOOL small_caps)
{	
    int fix = -1;
    char name[200];
    char *weight_str, *slant_str, *condensed_str, *serif_str, *rgstry;
    char *family;
    /*    int i; */
    HTList *l;

    px = (int)((double) px * lens_factor);
	
    /* if a point size is requested, let's compute the pixel size to
       avoid using the dpi in the request */

    switch (weight) {
      case SV_FONT_WEIGHT_LIGHT:
	weight_str = "light";
	break;
      case SV_FONT_WEIGHT_MEDIUM:
	weight_str = "medium";
	break;
      case SV_FONT_WEIGHT_DEMI_BOLD:
	weight_str = "demibold";
	break;
      case SV_FONT_WEIGHT_BOLD:
	weight_str = "bold";
	break;
      case SV_FONT_WEIGHT_EXTRA_BOLD:
	weight_str = "black";
	break;
      default:
	weight_str = "medium";
	break;	
    }

    switch (style) {
      case SV_FONT_STYLE_ROMAN:
	slant_str = "r";
	break;
      case SV_FONT_STYLE_ITALIC:
      case SV_FONT_STYLE_OBLIQUE:
	slant_str = "i";
	break;
      default:
	slant_str = "r";
	break;	
    }

    serif_str = "*";
    condensed_str = "normal";

/*    l = family_l; */

    while((fix < 0) && (family = (char *)family_l/*HTList_nextObject(l)*/)) {
	if (strcasecmp(family,"symbol") == 0)
	    rgstry = "*-*";
	else
	    rgstry = "iso8859-1";

	sprintf(name,"-*-%s-%s-%s-%s-%s-%ld-*-*-*-*-*-%s", family, weight_str, slant_str, condensed_str, serif_str, px, rgstry);
	fix = myLoadFont(name);

	if ((fix < 0) && ((style == SV_FONT_STYLE_ITALIC) || (style == SV_FONT_STYLE_OBLIQUE))){
	    slant_str = "o";
	    sprintf(name,"-*-%s-%s-%s-%s-%s-%ld-*-*-*-*-*-%s", family, weight_str, slant_str, condensed_str, serif_str, px, rgstry);
	    fix = myLoadFont(name);
	}

    }

    if (FONT_TRACE)
	fprintf(stderr,"font: %s %s\n", name,((fix >= 0) ? "succeded" : "failed, trying again"));

#if 0
    if ((fix < 0) && (style == SV_FONT_STYLE_ITALIC)) {
	slant_str = "o";

	l = family_l;
	while((fix <= 0) && (family = (char *)HTList_nextObject(l))) {
	    if (strcasecmp(family,"symbol") == 0)
		rgstry = "*-*";
	    else
		rgstry = "iso8859-1";

	    sprintf(name,"-*-%s-%s-%s-%s-%s-%ld-*-*-*-*-*-%s", family, weight_str, slant_str, condensed_str, serif_str, px, rgstry);
	    fix = myLoadFont(name);
	}

	if (FONT_TRACE)
	    fprintf(stderr,"font: %s %s\n", name,((fix >= 0) ? "succeded" : "failed, trying again"));
    }
#endif

    rgstry = "iso8859-1";

    if (fix < 0) {
	if (FONT_TRACE)
	    fprintf(stderr,"font: resorting to 9x15\n");
	strcpy(name,"9x15");
	fix = myLoadFont(name);
/*	fprintf(stderr,"font: %s %s\n", name,((fix >= 0) ? "succeded" : "failed, trying again"));*/
    }

    if (fix < 0) {
	if (FONT_TRACE)
	    fprintf(stderr,"font: resorting to 9x15\n");
	strcpy(name,"fixed");
	fix = myLoadFont(name);
/*	fprintf(stderr,"font: %s %s\n", name,((fix >= 0) ? "succeded" : "failed, trying again"));*/
    }

    if (fix > 255) {
	fprintf(stderr,"GetFont: WARNING! more that 255 fonts loaded!\n");
        Exit(0);
    }

    return fix;
}

float dpi(void)
{
    float wmm, w;

    wmm = DisplayWidthMM(display, screen);
    w   = DisplayWidth(display, screen);

    return((float)(w/wmm));
}

/*
   how to get from mm to px:

    1 inch = 25.4 mm = 72pt

    pix / mm ~~ 4 (100dpi)
                3 (75 dpi)

    14 pt / 2.83 (pt/mm)

    14 pt * mm/pt * pix/mm = pix
*/

#else  /* I18N */

#define	FONT_DB_SIZE	256		/* font index < 256 */
static struct {
    OutC	outc;
    HTList	*family_list;
    long	pixels;
    long	weight;
    long	style;
    BOOL	small_caps;
} font_db[FONT_DB_SIZE];
static int	font_db_index = 0;

#define	cmp_font_db(db, family_l, px, weight, style, scaps)	\
	(!(   px		== (db)->pixels			\
	   && weight		== (db)->weight			\
	   && style		== (db)->style			\
	   && small_caps	== (db)->small_caps		\
	   && family_list_cmp(family_l, font_db[i].family_list) == 0))

static OutCList gen_xlfd_list(HTList*, long, long, long, BOOL);
static OutCList add_xlfd(OutCList, int,
			 char*, char*, char*, char*, char*, char*,
			 int, int, int, int, char*, int, char*);

OutC fi2oc(int index)
{
    return font_db[index].outc;
}

int GetFont(char *family, long px, long weight, long style, BOOL small_caps)
{
    extern XFontStruct *Fonts[];	/* XXX */
    extern char *FontsName[];		/* XXX */
    extern int LineSpacing[], BaseLine[], StrikeLine[], LineThickness[];/*XXX*/
    extern double lens_size_factor;
    OutC outc;
    XFontStruct *fs;
    int i;
    HTList *family_l = HTList_new();

    HTList_addObject(family_l, strdup(family));

    if (FONT_TRACE) {
	printf("GetFont(family_l=%08x, px=%d, weight=%d, style=0x%08x, small_caps=%d\n",
	       family_l, px, weight, style, small_caps);
    }

    px = (int)((double) px * lens_factor);

#ifdef    I18N
    if (px == 15) px = 14;
    if (px == 18) px = 16;
    if (px == 22) px = 24;
    if (px == 23) px = 24;
    if (px == 28) px = 24;
    if (px > 36)  px = 36;
#endif /* I18N */

    /* search in font_db */
    for (i = 0; i < font_db_index; i++) {
	if (cmp_font_db(&font_db[i],
			family_l, px, weight, style, small_caps) == 0) {
	    if (FONT_TRACE) {
		printf("GetFont() found in cache (i=%i)\n", i);
	    }
	    return i;
	}
    }

    outc = CreateOutC(gen_xlfd_list(family_l, px, weight, style, small_caps));
    if (outc == NULL) {
	fprintf(stderr, "GetFont(): WARNING! failed to create OutC\n");
	Exit(0);
    }

    /* XXX: skip unappropriate indexes */
    for (i = font_db_index; i < FONT_DB_SIZE; i++) {
	if (Fonts[i] == NULL && FontsName[i] == NULL) {
	    font_db_index = i + 1;
	    break;
	}
    }
    if (i >= FONT_DB_SIZE) {
	fprintf(stderr, "GetFont(): WARNING! more than %d fonts loaded!\n",
		FONT_DB_SIZE - 1);
	Exit(0);
    }

    /* initialize font_db[i] */
    font_db[i].outc		= outc;
    font_db[i].family_list	= family_l;
    font_db[i].pixels		= px;
    font_db[i].weight		= weight;
    font_db[i].style		= style;
    font_db[i].small_caps	= small_caps;

    /* XXX: backward compatibility?? */
    fs = DefaultFontOfOutC(outc);
    Fonts[i]     	= fs;
    FontsName[i] 	= "no-name";		/* XXX */
    LineSpacing[i]	= SPACING(fs);
    BaseLine[i]		= BASELINE(fs);
    StrikeLine[i]	= STRIKELINE(fs);
    LineThickness[i]	= LINETHICKNESS(fs);

    if (FONT_TRACE)
	printf("GetFont() succeeded to load (i=%d)\n", i);
  
    return i;
}

static OutCList gen_xlfd_list(HTList *family_l, long px, long wt,
			      long style, BOOL small_caps)
{
    char *foundry, *family, *weight, *slant, *setwidth, *serif;
    char *spacing, *charset;
    int pixels, points, hres, vres, avarage, scaling;
    HTList *htl;
    OutCList list = NULL;

    foundry = "*";

    switch (wt) {
     case SV_FONT_WEIGHT_LIGHT:	weight = "light";	break;
     case SV_FONT_WEIGHT_MEDIUM:	weight = "medium";	break;
     case SV_FONT_WEIGHT_DEMI_BOLD:	weight = "demibold";	break;
     case SV_FONT_WEIGHT_BOLD:	weight = "bold";	break;
     case SV_FONT_WEIGHT_EXTRA_BOLD:	weight = "black";	break;
     default:			weight = "medium";	break;
    }

    switch (style) {
     case SV_FONT_STYLE_ROMAN:	slant  = "r";		break;
     case SV_FONT_STYLE_ITALIC:	slant  = "i";		break;
     default:			slant  = "r";		break;
    }

    setwidth = "normal";

#if 0
    switch (F_SERIF(style)) {
     case F_SERIF_SANS:		serif  = "sans";	break;
     case F_SERIF_AVEC:		serif  = "";		break;
     default:			serif  = "*";		break;
    }
#else
    serif = "*";
#endif

    pixels = px;
    points = hres = vres = avarage = -1;  /* `-1' means "*" */
    spacing = "*";
    charset = "*-*";
    scaling = 1;

    for (htl = family_l; family = (char*)HTList_nextObject(htl);) {
	list = add_xlfd(list, scaling, foundry, family, weight, 
			slant, setwidth, serif,
			pixels, points, hres, vres,
			spacing, avarage, charset);
	if (strcmp(slant, "i") == 0) {
	    list = add_xlfd(list, scaling, foundry, family, weight,
			    "o", setwidth, serif,
			    pixels, points, hres, vres,
			    spacing, avarage, charset);
	}
    }

    list = add_xlfd(list, scaling, foundry, "*", weight,
		    slant, setwidth, serif,
		    pixels, points, hres, vres,
		    spacing, avarage, charset);
    if (strcmp(slant, "i") == 0) {
	list = add_xlfd(list, scaling, foundry, "*", weight,
			"o", setwidth, serif,
			pixels, points, hres, vres,
			spacing, avarage, charset);
    }
    list = add_xlfd(list, scaling, "*", "*", "*", "*", "*", "*",
		    pixels, points, hres, vres,
		    spacing, avarage, charset);

    return list;
}

/* elements of XLFD
 * -foundry-family-weight-slant-setwidth-serif-\
 *  pixels-points-hres-vres-spacing-average-charset */
static OutCList add_xlfd(OutCList list, int scaling,
			 char* foundry, char* family, char* weight,
			 char* slant, char* setwidth, char* serif,
			 int pixels, int points, int hres, int vres,
			 char* spacing, int avarage, char* charset)
{
    char *xlfd, *scaling_str;
    char *points_str, *hres_str, *vres_str, *avarage_str;

    scaling_str = scaling ? "*-*-*" : "*";
    avarage_str = "*";				/* XXX */

    xlfd = malloc(256);   /* XXX: 256 would be enough for XLFD prefix */
    sprintf(xlfd, "-%s-%s-%s-%s-%s-%s-%d-%s-%s-%s-%s",
	    foundry, family, weight, slant, setwidth, serif,
	    pixels, scaling_str, spacing, avarage_str, charset);
    if (strlen(xlfd) >= 256) {
	printf("add_xlfd() WARNING!! xlfd(%s) is too long\n", xlfd);
    }
    xlfd = realloc(xlfd, strlen(xlfd)+1);
    return add_list(list, xlfd);
}

family_list_cmp(HTList* a, HTList* b)
{
    char *a_name, *b_name;
    int  ret;

    for (;;) {
	a_name = (char*)HTList_nextObject(a);
	b_name = (char*)HTList_nextObject(b);
	if (a_name == NULL && b_name == NULL)
	    break;
	if (a_name == NULL)
	    return -1;	/* b is longer than a. */
	if (b_name == NULL)
	    return 1;	/* a is longer than b. */
	ret = strcmp(a_name, b_name);
	if (ret)
	    return ret;	/* a_name and b_name aren't the same. */
    }
    return 0;	/* all entries are the same. */
}

/*
   HTAtom* HTAtom_for(CONST char* string);
   char* HTAtom_name(HTAtom* htatom);
 */

#endif /* I18N */


