#include <stdio.h>
#include "www_wchar.h"

#define BIG52CNS	1
#define CNS2BIG5	2


#define BIG52CNS_TABLE	"big5cns.table"

int mode;

unsigned char in[2048];
unsigned char out[3840];
uint16 table[0x6000];
int count;

main(argc, argv)
int argc;
char *argv[];
{
    char  *cp;

    mode = BIG52CNS;
    for (argc--,argv++; (argc > 0) && **argv == '-'; argc--, argv++) 
	{
	cp = *argv;
	while (*cp) 
	    {
	    switch (*cp++) {
	      case 'c':		/* conv to CNS */
		mode = BIG52CNS;
		continue;
	      case 'b':		/* conv to BIG5 */
		mode = CNS2BIG5;
		continue;
	      default:
		/* bogus option but ignored */
		continue;
		}
	    }
	}

    if (mode == BIG52CNS)
	{
	init_big5_table();
	while (fgets(in, sizeof(in), stdin) != NULL)
	    {
	    big52cns(in, out, 0);
	    dump_stdout(out);
	    }
	}
    else
	{
	init_cns_table();
	while (fgets(in, sizeof(in), stdin) != NULL)
	    {
	    cns2big5(in, out, 0);
	    dump_hex(out);
	    }
	}

    printf("\n");
}

dump_stdout(unsigned char *line)
{
unsigned char *data;
unsigned char hi, lo;

data = line;

while (*data != 0)
    {
    write(1, (char *)data, 1);
    data ++;
    }
}

dump_hex(uint16 *line)
{
uint16 *data;

data = line;

while (*data != 0)
    {
    if (((*data & 0xFF80) == 0) && isprint(*data))
	{
	printf("   %c ", *data);
	}
    else
	{
	printf("%04x ", *data);
	}

    data ++;
    count ++;
    if ((count % 8) == 0)
	printf("\n");
    }
}


/*
 * BIG5 to CNS
 */
int big52cns (unsigned char *in, unsigned char *out)
{
    static int state;
    unsigned char hi, lo;
    unsigned int val;
    wchar_t lang;

    state = S_single;
    lang = HEAD_CNS11643_1;

    while (*in != NULL)
	{
	switch (state) {
	  /* single byte character */
	  case S_single:
	    if (*in >= 0x80)
		{
		hi = *in ++;
		lo = *in ++;
		val = table[(hi << 8 | lo) - 0xa000];
		/* put escape sequence */
		*out ++ = '\033';
		*out ++ = '$';
		*out ++ = '(';
		if (val < 0x8000)
		    {
		    /* CNS-1 */
		    *out ++ = 'G';
		    lang = HEAD_CNS11643_1;
		    }
		else
		    {
		    /* CNS-2 */
		    *out ++ = 'H';
		    lang = HEAD_CNS11643_2;
		    }
		state = S_multi;
		*out ++ = (val & 0x7f00) >> 8;
		*out ++ = (val & 0x00ff);
		}
	    else
		{
		*out ++ = *in ++;
		}
	    break;

	  /* multi byte character */
	  case S_multi:
	    if (*in < 0x80)
		{
		/* put escape sequence to ISO-646 IRV */
		*out ++ = '\033';
		*out ++ = '(';
		*out ++ = 'B'; /* XXX ASCII */
		state = S_single;
		*out ++ = *in ++;
		}
	    else
		{
		hi = *in;
		lo = *(in + 1);
		val = table[(hi << 8 | lo) - 0xa000];

		if ((val < 0x8000) && (lang == HEAD_CNS11643_2))
		    {
		    /* CNS-2 -> 1 */
		    state = S_single;
		    break;
		    }
		else if ((val >= 0x8000) && (lang == HEAD_CNS11643_1))
		    {
		    /* CNS-1 -> 2 */
		    state = S_single;
		    break;
		    }
		else
		    {
		    *out ++ = (val & 0x7f00) >> 8;
		    *out ++ = (val & 0x00ff);
		    in ++;
		    in ++;
		    }
		}
	    }
	}
    *out ++ = NULL;	/* NULL termination */
}



/*
 * BIG5 to CNS
 */
int cns2big5 (unsigned char *in, unsigned char *out)
{
}



/*
 * Make conv table BIG5 to CNS-1,2
 */
init_big5_table()
{
FILE *in;
unsigned int big5, plane, cns;
unsigned int i1, i2, i3;

/* open conv table */
in = fopen(BIG52CNS_TABLE, "r");
if (in == NULL)
    {
    perror("conversion table");
    exit(-1);
    }

while (fscanf(in, "%x\t%x-%x", &big5, &plane, &cns) != EOF)
    {
    if (plane == 1)
	{
	table[big5 - 0xa000] = cns;
	}
    else
	{
	/* MSB=1 indicates CNS-2 */
	table[big5 - 0xa000] = (cns | 0x8000);
	}
    }
fclose(in);
}

init_cns_table()
{
}
