/*-
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Chris Torek.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "www_wchar.h"



IS_ALPHA (wchar_t c)
    {
    return (IS_ASCII(c) && isalpha((char)c));
    }


IS_DIGIT (wchar_t c)
    {
    return (IS_ASCII(c) && isdigit((char)c));
    }

IS_ALNUM (wchar_t c)
    {
    return (IS_ASCII(c) && (isdigit((char)c) || isalpha((char)c)));
    }

IS_LOWER (wchar_t c)
    {
    return (IS_ASCII(c) && islower((char)c));
    }

wchar_t
W_tolower(wchar_t c)
    {
    return ((c & MASK_LATIN_1) | (wchar_t)tolower((char)(c & ~MASK_LATIN_1)));
    }

wchar_t
W_toupper(wchar_t c)
    {
    return ((c & MASK_LATIN_1) | (wchar_t)toupper((char)(c & ~MASK_LATIN_1)));
    }


/*
 * This array is designed for mapping upper and lower case letter
 * together for a case independent comparison.  The mappings are
 * based upon ascii character sequences.
 */
static unsigned char charmap[] = {
	'\000', '\001', '\002', '\003', '\004', '\005', '\006', '\007',
	'\010', '\011', '\012', '\013', '\014', '\015', '\016', '\017',
	'\020', '\021', '\022', '\023', '\024', '\025', '\026', '\027',
	'\030', '\031', '\032', '\033', '\034', '\035', '\036', '\037',
	'\040', '\041', '\042', '\043', '\044', '\045', '\046', '\047',
	'\050', '\051', '\052', '\053', '\054', '\055', '\056', '\057',
	'\060', '\061', '\062', '\063', '\064', '\065', '\066', '\067',
	'\070', '\071', '\072', '\073', '\074', '\075', '\076', '\077',
	'\100', '\141', '\142', '\143', '\144', '\145', '\146', '\147',
	'\150', '\151', '\152', '\153', '\154', '\155', '\156', '\157',
	'\160', '\161', '\162', '\163', '\164', '\165', '\166', '\167',
	'\170', '\171', '\172', '\133', '\134', '\135', '\136', '\137',
	'\140', '\141', '\142', '\143', '\144', '\145', '\146', '\147',
	'\150', '\151', '\152', '\153', '\154', '\155', '\156', '\157',
	'\160', '\161', '\162', '\163', '\164', '\165', '\166', '\167',
	'\170', '\171', '\172', '\173', '\174', '\175', '\176', '\177',
	'\200', '\201', '\202', '\203', '\204', '\205', '\206', '\207',
	'\210', '\211', '\212', '\213', '\214', '\215', '\216', '\217',
	'\220', '\221', '\222', '\223', '\224', '\225', '\226', '\227',
	'\230', '\231', '\232', '\233', '\234', '\235', '\236', '\237',
	'\240', '\241', '\242', '\243', '\244', '\245', '\246', '\247',
	'\250', '\251', '\252', '\253', '\254', '\255', '\256', '\257',
	'\260', '\261', '\262', '\263', '\264', '\265', '\266', '\267',
	'\270', '\271', '\272', '\273', '\274', '\275', '\276', '\277',
	'\300', '\301', '\302', '\303', '\304', '\305', '\306', '\307',
	'\310', '\311', '\312', '\313', '\314', '\315', '\316', '\317',
	'\320', '\321', '\322', '\323', '\324', '\325', '\326', '\327',
	'\330', '\331', '\332', '\333', '\334', '\335', '\336', '\337',
	'\340', '\341', '\342', '\343', '\344', '\345', '\346', '\347',
	'\350', '\351', '\352', '\353', '\354', '\355', '\356', '\357',
	'\360', '\361', '\362', '\363', '\364', '\365', '\366', '\367',
	'\370', '\371', '\372', '\373', '\374', '\375', '\376', '\377',
};


int
W_strcasecmp(s1, s2)
    wchar_t *s1;
    char *s2;
{
	unsigned char *cm = charmap;
	wchar_t *us1 = (wchar_t *)s1;
	char *us2 = (unsigned char *)s2;

	while (IS_ASCII(*us1) && cm[*us1] == cm[*us2++])
		if (*us1++ == '\0')
			return (0);
	return (*us1 - (wchar_t)(*--us2));
}


/*
 * s1: wchar
 * s2: char
 * len: length of strings.
 * return: 
 *	0     = same.
 * 	other = unknown.
 */

int
W_strncasecmp(s1, s2, len)
wchar_t *s1;
unsigned char * s2;
int len;
    {
    unsigned char *cm = charmap;
    wchar_t *us1 = s1;
    unsigned char *us2 = s2;
    if (len != 0) 
	    {
	    
	    do 
		{
		if (!IS_ASCII(*us1) || (cm[*us1] != cm[*us2++]))
		    return (*us1 - (wchar_t)(*--us2));
		if (*us1++ == '\0')
		    break;
		} while (--len != 0);
	    }
	return (0);
    }


/* duplicate  string s and ensure it is zero terminated
   where n is the maximum number of chars copied from s */

char *
W_strndup(wchar_t *s, int n)
{
    char *p;
    int i;

    p = (char *)malloc(n+1);

    if (!p)
    {
        Beep();
        fprintf(stderr, "ERROR: malloc failed, memory exhausted!\n");
        exit(1);
    }

    for(i = 0; i < n; i++)
	{
	p[i] = (char)(*s++ & ~MASK_LATIN_1);
	}
    p[n] = '\0';

    return p;
}


size_t
W_strlen(str)
	wchar_t *str;
{
	wchar_t *s;

	for (s = str; *s; ++s);
	return(s - str);
}


char *
W_strdup(wchar_t *s)
{
    char *t;
    int i, len;

    if (s == NULL)
	return NULL;
#if 0
    t = (char*)malloc((W_strlen(s)+1) * 6);
    if (t == NULL)
	return NULL;
    wchar2mb(s, t, 0, "iso-2022-int-1");	/* XXX */
    t = realloc(t, strlen(t)+1);
#else
    len = W_strlen(s);
    t = (char*)malloc(len+1);
    if (t == NULL)
	return NULL;
    for (i = 0; i < len; i++) {
	t[i] = (char)(*s++ & ~MASK_LATIN_1);
    }
#endif
    return t;
}

/*
 * string duplicate --- wchar_t to wchar_t
 */
wchar_t *
WW_strdup(wchar_t *s)
{
    wchar_t *t;
    int i, len;

    if (s == NULL)
	return NULL;
    len = W_strlen(s);
    t = (wchar_t *)malloc(sizeof(wchar_t)*(len + 1));
    if (t == NULL)
	return NULL;
    for (i = 0; i < len; i++) {
	t[i] = *s++;
    }
    t[i + 1] = 0;
    return t;
}


wchar_t *
w_strdup(char* s)
{
    wchar_t *t;

    if (s == NULL)
	return NULL;
    t = (wchar_t*)malloc((strlen(s)+1) * sizeof(wchar_t));
    mb2wchar(s, t, 0, "iso-2022-int-1");
    return t;
}


wchar_t *
w_strndup(char* s, int len)
{
    wchar_t *t;

    if (s == NULL)
	return NULL;
    t = (wchar_t*)malloc((len+1) * sizeof(wchar_t));
    mb2wchar(s, t, 0, "iso-2022-int-1");	/* XXX */
    return t;
}


int
W_strcmp(s1, s2)
wchar_t* s1;
char*    s2;
{
    while (*s1) {
	if ( !IS_LATIN_1(*s1) || (TO_ASCII(*s1) != *s2) )
	    return *s1 - *s2;
	s1++;
	s2++;
    }
    return *s2;
}


int
W_strncmp(s1, s2, n)
     wchar_t *s1;
     char *s2;
     size_t n;
{

	if (n == 0)
		return (0);
	do {
		if (*s1 != (wchar_t)(*s2++))
			return (*s1 - (wchar_t)(*--s2));
		if (*s1++ == 0)
			break;
	} while (--n != 0);
	return (0);
}


/*
 * Find the first occurrence of find in s.
 */
wchar_t *
W_strstr(s, find)
     wchar_t *s;
     char *find;
{
    char c;
    wchar_t sc;
    size_t len;

	if ((c = *find++) != 0) {
		len = strlen(find);
		do {
			do {
				if ((sc = *s++) == 0)
					return (NULL);
			} while (sc != (wchar_t)c);
		} while (W_strncmp(s, find, len) != 0);
		s--;
	}
	return ((wchar_t *)s);
}

wchar_t *
W_strchr(wchar_t *str, char c)
{
    while (*str) {
	if (*str == (wchar_t)c)
	    return str;
    }
    return (wchar_t*)NULL;
}

int
W_atoi(wchar_t *str)
    {
    int c;
    char *s = W_strndup(str, 16);

    c = atoi(s);
    free(s);
    return c;
    }

int
STR_NCASECMP(s1, s2, len)
wchar_t *s1, *s2;
int len;
{
    if (len != 0) {
	unsigned char *cm = charmap;
	wchar_t *us1 = s1;
	wchar_t *us2 = s2;

	do {
	    if (!IS_ASCII(*us1) || !IS_ASCII(*us2) || cm[*us1] != cm[*us2++])
		return (*us1 - *--us2);
	    if (*us1++ == '\0')
		break;
	    } while (--len != 0);
	}
    return(0);
}
